package presence

import (
	"context"
	"log"
	"math/rand"
	"sync"

	"github.com/golang/protobuf/proto"
	"go.etcd.io/etcd/clientv3"

	"git.autistici.org/ale/autoradio/coordination/watcher"
	pb "git.autistici.org/ale/autoradio/proto"
)

// EndpointSet is a container of Endpoints that is synchronizable with
// the contents of a presence tree as created by Register().
type EndpointSet struct {
	mx sync.Mutex

	// Endpoints by etcd key.
	eps map[string]*pb.Endpoint

	// Endpoints by name.
	epn map[string]*pb.Endpoint

	// List of endpoints for fast RandomEndpoint().
	epl []*pb.Endpoint
}

// Reset implements the watcher.Watchable interface.
func (n *EndpointSet) Reset(m map[string]string) {
	eps := make(map[string]*pb.Endpoint)
	epn := make(map[string]*pb.Endpoint)
	var epl []*pb.Endpoint
	for k, v := range m {
		var ep pb.Endpoint
		if err := proto.Unmarshal([]byte(v), &ep); err != nil {
			log.Printf("presence: error unmarshaling %s: %v", k, err)
			continue
		}
		eps[k] = &ep
		epn[ep.Name] = &ep
		epl = append(epl, &ep)
	}
	n.mx.Lock()
	n.eps = eps
	n.epn = epn
	n.epl = epl
	n.mx.Unlock()
}

// Set implements the watcher.Watchable interface.
func (n *EndpointSet) Set(k, v string) {
	var ep pb.Endpoint
	if err := proto.Unmarshal([]byte(v), &ep); err != nil {
		log.Printf("presence: error unmarshaling %s: %v", k, err)
		return
	}
	n.mx.Lock()
	if old, ok := n.eps[k]; ok {
		n.epl = withoutEndpoint(n.epl, old)
	}
	n.epl = append(n.epl, &ep)
	if n.eps == nil {
		n.eps = make(map[string]*pb.Endpoint)
	}
	if n.epn == nil {
		n.epn = make(map[string]*pb.Endpoint)
	}
	n.eps[k] = &ep
	n.epn[ep.Name] = &ep
	n.mx.Unlock()
}

// Delete implements the watcher.Watchable interface.
func (n *EndpointSet) Delete(k string) {
	n.mx.Lock()
	defer n.mx.Unlock()
	if n.eps == nil {
		return
	}
	if old, ok := n.eps[k]; ok {
		delete(n.eps, k)
		delete(n.epn, old.Name)
		n.epl = withoutEndpoint(n.epl, old)
	}
}

func withoutEndpoint(l []*pb.Endpoint, old *pb.Endpoint) []*pb.Endpoint {
	out := make([]*pb.Endpoint, 0, len(l))
	for _, ep := range l {
		if ep != old {
			out = append(out, ep)
		}
	}
	return out
}

// Endpoints returns all registered endpoints.
func (n *EndpointSet) Endpoints() []*pb.Endpoint {
	n.mx.Lock()
	defer n.mx.Unlock()
	if len(n.epl) == 0 {
		return nil
	}
	return append([]*pb.Endpoint{}, n.epl...)
}

// RandomEndpoint returns a randomly selected Endpoint (or nil, if
// none are available).
func (n *EndpointSet) RandomEndpoint() *pb.Endpoint {
	n.mx.Lock()
	defer n.mx.Unlock()
	if len(n.epl) == 0 {
		return nil
	}
	return n.epl[rand.Intn(len(n.epl))]
}

// RandomEndpointExcluding returns a randomly selected Endpoint except
// for the named one (or nil, if none are available).
func (n *EndpointSet) RandomEndpointExcluding(name string) *pb.Endpoint {
	n.mx.Lock()
	defer n.mx.Unlock()
	l := len(n.epl)
	if l == 0 {
		return nil
	}
	// Find the excluded element.
	found := -1
	for i := 0; i < l; i++ {
		if n.epl[i].Name == name {
			found = i
			break
		}
	}
	// Is the excluded element not in the list?
	if found < 0 {
		return n.epl[rand.Intn(l)]
	}
	// Is the list empty once we exclude the item?
	if l == 1 {
		return nil
	}
	i := rand.Intn(l - 1)
	if i >= found {
		i++
	}
	return n.epl[i]
}

// GetEndpoint returns a specific endpoint by host name.
func (n *EndpointSet) GetEndpoint(name string) (*pb.Endpoint, bool) {
	n.mx.Lock()
	defer n.mx.Unlock()
	ep, ok := n.epn[name]
	return ep, ok
}

func WatchEndpoints(ctx context.Context, cli *clientv3.Client, prefix string) *EndpointSet {
	var endpoints EndpointSet
	go watcher.Watch(ctx, cli, prefix, &endpoints)
	return &endpoints
}
