package watcher

import (
	"context"
	"fmt"
	"log"
	"math/rand"
	"sync"
	"time"

	"go.etcd.io/etcd/clientv3"
)

var watcherErrDelay = 200 * time.Millisecond

// A Watchable exposes an interface to allow synchronization with the
// remote etcd contents.
type Watchable interface {
	Reset(map[string]string)
	Set(string, string)
	Delete(string)
}

// A ReadyWatchable is a Watchable that exposes a method to let you
// know when it has the initial complete view of the data.
type ReadyWatchable interface {
	Watchable
	WaitForInit() <-chan struct{}
}

// Implementation of the above.
type readyChWatchable struct {
	Watchable

	ready   sync.Once
	readyCh chan struct{}
}

// Reset the contents of the wrapped Watchable. It will trigger the
// WaitForInit channel the first time it is called.
func (r *readyChWatchable) Reset(m map[string]string) {
	r.Watchable.Reset(m)
	r.ready.Do(func() {
		close(r.readyCh)
	})
}

func (r *readyChWatchable) WaitForInit() <-chan struct{} {
	return r.readyCh
}

// NewReadyWatchable wraps a Watchable with a WaitForInit method.
func NewReadyWatchable(w Watchable) ReadyWatchable {
	return &readyChWatchable{
		Watchable: w,
		readyCh:   make(chan struct{}),
	}
}

// A NotifyWatchable is a Watchable with a channel to receive updates
// whenever the data changes.
type NotifyWatchable interface {
	Watchable
	Notify() <-chan struct{}
}

// Implementation of the above.
type notifyChWatchable struct {
	Watchable
	notifyCh chan struct{}
}

func (n *notifyChWatchable) Notify() <-chan struct{} {
	return n.notifyCh
}

func (n *notifyChWatchable) awaken() {
	select {
	case n.notifyCh <- struct{}{}:
	default:
	}
}

func (n *notifyChWatchable) Reset(m map[string]string) {
	n.Watchable.Reset(m)
	n.awaken()
}

func (n *notifyChWatchable) Set(k, v string) {
	n.Watchable.Set(k, v)
	n.awaken()
}

func (n *notifyChWatchable) Delete(k string) {
	n.Watchable.Delete(k)
	n.awaken()
}

// NewNotifyWatchable returns a NotifyWatchable wrapping another
// Watchable.
func NewNotifyWatchable(ctx context.Context, w Watchable) NotifyWatchable {
	notifyCh := make(chan struct{}, 1)
	// Close the channel when the context is done.
	go func() {
		<-ctx.Done()
		close(notifyCh)
	}()
	return &notifyChWatchable{
		Watchable: w,
		notifyCh:  notifyCh,
	}
}

// A StringSet is a particular type of Watchable where we are only
// interested in the entry values, not their keys. Furthermore, the
// values are plain strings. This is useful, for instance, for a list
// of IP addresses.
//
// The StringSet keeps track of the keys, so if an entry changes value
// we can handle it correctly (the previous value is removed from the
// set).
//
// All methods are goroutine-safe. The nil value represents an empty
// but valid set.
//
type StringSet struct {
	mx sync.Mutex
	m  map[string]string
	l  []string
}

// Reset the contents of the set. Implements the Watchable interface.
func (s *StringSet) Reset(m map[string]string) {
	s.mx.Lock()
	s.m = m
	s.l = nil
	for _, value := range m {
		s.l = append(s.l, value)
	}
	s.mx.Unlock()
}

// Add an entry to the set. Implements the Watchable interface.
func (s *StringSet) Add(key, value string) {
	s.mx.Lock()
	if s.m == nil {
		s.m = make(map[string]string)
	}
	oldValue, ok := s.m[key]
	if ok {
		s.deleteValue(oldValue)
	}
	s.m[key] = value
	s.l = append(s.l, value)
}

// Delete an entry from the set. Implements the Watchable interface.
func (s *StringSet) Delete(key string) {
	s.mx.Lock()
	defer s.mx.Unlock()

	if s.m == nil {
		return
	}
	oldValue, ok := s.m[key]
	if !ok {
		return
	}
	s.deleteValue(oldValue)
	delete(s.m, key)
}

func (s *StringSet) deleteValue(oldValue string) {
	var l []string
	for _, value := range s.l {
		if value != oldValue {
			l = append(l, value)
		}
	}
	s.l = l
}

// RandomEntry returns a random entry from those currently active. If
// the list is empty, the empty string is returned.
func (s *StringSet) RandomEntry() string {
	s.mx.Lock()
	defer s.mx.Unlock()
	if len(s.l) == 0 {
		return ""
	}
	return s.l[rand.Intn(len(s.l))]
}

// Values returns all the currently known values.
func (s *StringSet) Values() []string {
	s.mx.Lock()
	defer s.mx.Unlock()
	// Try to make a copy.
	return append([]string{}, s.l...)
}

func watchOnce(ctx context.Context, cli *clientv3.Client, prefix string, target Watchable) error {
	// First populate the target with a prefix-ranged Get and call
	// Reset because we have all the authoritative data.
	resp, err := cli.Get(ctx, prefix, clientv3.WithPrefix(), clientv3.WithSort(clientv3.SortByKey, clientv3.SortAscend))
	if err != nil {
		return err
	}
	m := make(map[string]string)
	for _, ev := range resp.Kvs {
		m[string(ev.Key)] = string(ev.Value)
	}
	target.Reset(m)
	rev := resp.Header.Revision

	// Then start the Watcher on the same revision.
	rch := cli.Watch(ctx, prefix, clientv3.WithPrefix(), clientv3.WithRev(rev))
	for resp := range rch {
		for _, ev := range resp.Events {
			fmt.Printf("%s %q : %q\n", ev.Type, ev.Kv.Key, ev.Kv.Value)
			switch ev.Type {
			case clientv3.EventTypePut:
				target.Set(string(ev.Kv.Key), string(ev.Kv.Value))
			case clientv3.EventTypeDelete:
				target.Delete(string(ev.Kv.Key))
			}
		}
	}

	// If we get here, the Watcher has been somehow disrupted, but
	// we don't have a specific error to return.
	return nil
}

// Watch a prefix on etcd and synchronize its contents with
// target. The function will return when the context is canceled,
// otherwise it will continuously attempt to re-establish the
// underlying Watcher on error.
func Watch(ctx context.Context, cli *clientv3.Client, prefix string, target Watchable) {
	for {
		err := watchOnce(ctx, cli, prefix, target)
		if err == context.Canceled {
			return
		} else if err != nil {
			log.Printf("watcher error: %s: %v", prefix, err)
		}

		// Wait a bit, then retry.
		time.Sleep(watcherErrDelay)
	}
}
