package node

import (
	"fmt"
	"log"
	"net"
	"strconv"
	"sync"
)

// A genericServer is just something that can be started and stopped.
type genericServer interface {
	Name() string
	Serve() error
	Stop()
}

// The Server runs all the request-based components of a Node. It
// bundles together servers for all the supported protocols (HTTP,
// DNS, GRPC). A failure of any of them will cause the entire Server
// to fail.
type Server struct {
	wg     sync.WaitGroup
	stopCh chan struct{}
	errCh  chan error
}

func buildServer(servers ...genericServer) *Server {
	ms := Server{
		stopCh: make(chan struct{}),
		errCh:  make(chan error, 1),
	}

	for _, s := range servers {
		ms.wg.Add(1)
		go func(s genericServer) {
			defer ms.wg.Done()
			log.Printf("starting network service: %s", s.Name())
			err := s.Serve()
			if err != nil {
				werr := fmt.Errorf("%s: %v", s.Name(), err)
				select {
				case ms.errCh <- werr:
					// First error, stop all other services.
					ms.Stop()
				default:
				}
			}
		}(s)
		go func(s genericServer) {
			<-ms.stopCh
			s.Stop()
		}(s)
	}

	return &ms
}

// Stop all network services.
func (s *Server) Stop() {
	close(s.stopCh)
}

// Wait for the services to terminate.
func (s *Server) Wait() error {
	s.wg.Wait()

	// There may or may not be an error stored in errCh.
	var err error
	select {
	case err = <-s.errCh:
	default:
	}
	close(s.errCh)

	return err
}

// NewServer creates a new Server. Will use publicAddrs / peerAddr to
// build all the necessary addr/port combinations.
func NewServer(n *Node, domain string, nameservers []string, publicAddrs []net.IP, peerAddr net.IP, httpPort, dnsPort, gossipPort, icecastPort, metricsPort int) *Server {
	httpHandler := newHTTPHandler(n, icecastPort, domain)
	dnsHandler := newDNSHandler(n, domain, nameservers)

	servers := []genericServer{
		newStatusServer(mkaddr(peerAddr, gossipPort), n.statusMgr),
		newHTTPServer("metrics", fmt.Sprintf(":%d", metricsPort), newMetricsHandler()),
	}
	for _, ip := range publicAddrs {
		servers = append(servers,
			newHTTPServer("main", mkaddr(ip, httpPort), httpHandler),
			newDNSServer("dns(udp)", mkaddr(ip, dnsPort), "udp", dnsHandler),
			newDNSServer("dns(tcp)", mkaddr(ip, dnsPort), "tcp", dnsHandler),
		)
	}
	return buildServer(servers...)
}

func mkaddr(ip net.IP, port int) string {
	return net.JoinHostPort(ip.String(), strconv.Itoa(port))
}
